#include <linux/kernel.h>
#include <linux/string.h>
#include "sms4-core.h"

static u32 __rol32(u32 word, unsigned int shift)
{
	return (word << shift) | (word >> (32 - shift));
}

/* Sbox table: 8bits input convert to 8 bits output*/
static u8 sms4_sbox[256] = {
	0xd6, 0x90, 0xe9, 0xfe, 0xcc, 0xe1, 0x3d, 0xb7,
	0x16, 0xb6, 0x14, 0xc2, 0x28, 0xfb, 0x2c, 0x05,
	0x2b, 0x67, 0x9a, 0x76, 0x2a, 0xbe, 0x04, 0xc3,
	0xaa, 0x44, 0x13, 0x26, 0x49, 0x86, 0x06, 0x99,
	0x9c, 0x42, 0x50, 0xf4, 0x91, 0xef, 0x98, 0x7a,
	0x33, 0x54, 0x0b, 0x43, 0xed, 0xcf, 0xac, 0x62,
	0xe4, 0xb3, 0x1c, 0xa9, 0xc9, 0x08, 0xe8, 0x95,
	0x80, 0xdf, 0x94, 0xfa, 0x75, 0x8f, 0x3f, 0xa6,
	0x47, 0x07, 0xa7, 0xfc, 0xf3, 0x73, 0x17, 0xba,
	0x83, 0x59, 0x3c, 0x19, 0xe6, 0x85, 0x4f, 0xa8,
	0x68, 0x6b, 0x81, 0xb2, 0x71, 0x64, 0xda, 0x8b,
	0xf8, 0xeb, 0x0f, 0x4b, 0x70, 0x56, 0x9d, 0x35,
	0x1e, 0x24, 0x0e, 0x5e, 0x63, 0x58, 0xd1, 0xa2,
	0x25, 0x22, 0x7c, 0x3b, 0x01, 0x21, 0x78, 0x87,
	0xd4, 0x00, 0x46, 0x57, 0x9f, 0xd3, 0x27, 0x52,
	0x4c, 0x36, 0x02, 0xe7, 0xa0, 0xc4, 0xc8, 0x9e,
	0xea, 0xbf, 0x8a, 0xd2, 0x40, 0xc7, 0x38, 0xb5,
	0xa3, 0xf7, 0xf2, 0xce, 0xf9, 0x61, 0x15, 0xa1,
	0xe0, 0xae, 0x5d, 0xa4, 0x9b, 0x34, 0x1a, 0x55,
	0xad, 0x93, 0x32, 0x30, 0xf5, 0x8c, 0xb1, 0xe3,
	0x1d, 0xf6, 0xe2, 0x2e, 0x82, 0x66, 0xca, 0x60,
	0xc0, 0x29, 0x23, 0xab, 0x0d, 0x53, 0x4e, 0x6f,
	0xd5, 0xdb, 0x37, 0x45, 0xde, 0xfd, 0x8e, 0x2f,
	0x03, 0xff, 0x6a, 0x72, 0x6d, 0x6c, 0x5b, 0x51,
	0x8d, 0x1b, 0xaf, 0x92, 0xbb, 0xdd, 0xbc, 0x7f,
	0x11, 0xd9, 0x5c, 0x41, 0x1f, 0x10, 0x5a, 0xd8,
	0x0a, 0xc1, 0x31, 0x88, 0xa5, 0xcd, 0x7b, 0xbd,
	0x2d, 0x74, 0xd0, 0x12, 0xb8, 0xe5, 0xb4, 0xb0,
	0x89, 0x69, 0x97, 0x4a, 0x0c, 0x96, 0x77, 0x7e,
	0x65, 0xb9, 0xf1, 0x09, 0xc5, 0x6e, 0xc6, 0x84,
	0x18, 0xf0, 0x7d, 0xec, 0x3a, 0xdc, 0x4d, 0x20,
	0x79, 0xee, 0x5f, 0x3e, 0xd7, 0xcb, 0x39, 0x48
};

static const u32 FK[4] = { 0xa3b1bac6, 0x56aa3350, 0x677d9197, 0xb27022dc };

static const u32 CK[32] = {
	0x00070e15, 0x1c232a31, 0x383f464d, 0x545b6269,
	0x70777e85, 0x8c939aa1, 0xa8afb6bd, 0xc4cbd2d9,
	0xe0e7eef5, 0xfc030a11, 0x181f262d, 0x343b4249,
	0x50575e65, 0x6c737a81, 0x888f969d, 0xa4abb2b9,
	0xc0c7ced5, 0xdce3eaf1, 0xf8ff060d, 0x141b2229,
	0x30373e45, 0x4c535a61, 0x686f767d, 0x848b9299,
	0xa0a7aeb5, 0xbcc3cad1, 0xd8dfe6ed, 0xf4fb0209,
	0x10171e25, 0x2c333a41, 0x484f565d, 0x646b7279
};

static u32 sms4_tau(u32 a)
{
	return sms4_sbox[a & 0xFF] |
		(sms4_sbox[(a >> 8) & 0xFF] << 8) |
		(sms4_sbox[(a >> 16) & 0xFF] << 16) |
		(sms4_sbox[(a >> 24) & 0xFF] << 24);
}

static u32 sms4_L(u32 a)
{
	return a ^ __rol32(a, 2) ^ __rol32(a, 10)
		^ __rol32(a, 18) ^ __rol32(a, 24);
}

/**
 * sms4_T - calculate T
 * @a: input
 *
 * Calculate the mixer substitution T(a).
 * T(.) = L(tau(.)), where tau(.) is the sbox
 * application on each octet of a 32-bit word,
 * and L is a given linear substitution.
 */
static u32 sms4_T(u32 a)
{
	return sms4_L(sms4_tau(a));
}

/**
 * sms4_Tprime - calculate T'
 * @a: input
 *
 * Calculate the mixer substitution T'(a).
 * T'(.) = L'(tau(.)), where tau(.) is the sbox
 * application on each octet of a 32-bit word,
 * and L is a given linear substitution.
 */
static u32 sms4_Tprime(u32 a)
{
	u32 b = sms4_tau(a);

	return b ^ __rol32(b, 13) ^ __rol32(b, 23);
}

/**
 * sms4_F - calculate F
 * @x0: original contents
 * @x1: ditto
 * @x2: ditto
 * @x3: ditto
 * @rk: round key
 */
static u32 sms4_F(u32 x0, u32 x1, u32 x2, u32 x3, u32 rk)
{
	return x0 ^ sms4_T(x1 ^ x2 ^ x3 ^ rk);
}

/**
 * sms4_crypt - SMS4 encryption/decryption routine
 *
 * @plain: input buffer
 * @out: output buffer
 * @rk: round key (32 words)
 */
static void sms4_crypt(const void *in, void *out, const u32 *rk)
{
	u32 tmp[36];
	u16 i;
	const __be32 *_in = in;
	__be32 *_out = out;

	/* copy input, padded with zeroes */
	tmp[0] = be32_to_cpu(_in[0]);
	tmp[1] = be32_to_cpu(_in[1]);
	tmp[2] = be32_to_cpu(_in[2]);
	tmp[3] = be32_to_cpu(_in[3]);
	memset(tmp + 4, 0, 32 * sizeof(u32));

	for (i = 0; i < 32; i++)
		tmp[i + 4] = sms4_F(tmp[i], tmp[i + 1],
				    tmp[i + 2], tmp[i + 3],
				    rk[i]);

	/* save encrypted contents to original area */
	_out[0] = cpu_to_be32(tmp[35]);
	_out[1] = cpu_to_be32(tmp[34]);
	_out[2] = cpu_to_be32(tmp[33]);
	_out[3] = cpu_to_be32(tmp[32]);
}

void sms4_set_key(struct sms4_ctx *ctx, const u8 *in_key)
{
	const __be32 *_mk = (void *)in_key;
	u32 k[36];
	u16 i;

	k[0] = be32_to_cpu(_mk[0]) ^ FK[0];
	k[1] = be32_to_cpu(_mk[1]) ^ FK[1];
	k[2] = be32_to_cpu(_mk[2]) ^ FK[2];
	k[3] = be32_to_cpu(_mk[3]) ^ FK[3];

	for (i = 0; i < 32; i++) {
		k[i + 4] = k[i] ^
			sms4_Tprime(k[i + 1] ^ k[i + 2] ^ k[i + 3] ^ CK[i]);
		ctx->rk_dec[31 - i] = k[i + 4];
		ctx->rk_enc[i] = k[i + 4];
	}
}

void sms4_encrypt(const struct sms4_ctx *ctx, u8 *out, const u8 *in)
{
	sms4_crypt(in, out, ctx->rk_enc);
}

void sms4_decrypt(const struct sms4_ctx *ctx, u8 *out, const u8 *in)
{
	sms4_crypt(in, out, ctx->rk_dec);
}

static void xor_block(u8 *target, const u8 *operand)
{
	u16 i;
	for (i = 0; i < SMS4_BLOCK_SIZE; i++)
		target[i] ^= operand[i];
}

static void swap_block(u8 *dest, const u8 *src, size_t size)
{
	unsigned int i;
	for (i = 0; i < size; i++)
		dest[i] = src[size - i - 1];
}

void sms4_crypt_buf_by_ofb(struct sms4_ctx *ctx, const u8 *iv_le,
	u8 *buf, size_t buf_size)
{
	u8 out[SMS4_BLOCK_SIZE] = {0x0};
	u8 in[SMS4_BLOCK_SIZE] = {0x0};
	u8 iv[SMS4_BLOCK_SIZE] = {0x0};
	u16 last_block_size;
	u16 i;

	swap_block(iv, iv_le, SMS4_BLOCK_SIZE);
	memcpy(in, iv, SMS4_BLOCK_SIZE);
	for (i = 0; i + SMS4_BLOCK_SIZE < buf_size; i += SMS4_BLOCK_SIZE) {
		sms4_encrypt(ctx, out, in);
		xor_block(buf + i, out);
		memcpy(in, out, SMS4_BLOCK_SIZE);
	}
	/* special treatment -
	   last block might be shorter than SMS4_BLOCK_SIZE */
	last_block_size = buf_size % SMS4_BLOCK_SIZE;
	if (last_block_size == 0)
		last_block_size = SMS4_BLOCK_SIZE;
	sms4_encrypt(ctx, out, in);
	memset(in, 0, SMS4_BLOCK_SIZE);
	memcpy(in, buf + i, last_block_size);
	xor_block(in, out);
	memcpy(buf + i, in, last_block_size);
}

void sms4_calculate_mic_by_cbc_mac(struct sms4_ctx *ctx, const u8 *iv_le,
	const u8 *integrity_check_data, size_t icd_size, u8 *mic)
{
	u8 out[SMS4_BLOCK_SIZE] = {0x0};
	u8 in[SMS4_BLOCK_SIZE] = {0x0};
	u8 iv[SMS4_BLOCK_SIZE] = {0x0};
	u16 i;

	swap_block(iv, iv_le, SMS4_BLOCK_SIZE);
	memcpy(in, iv, SMS4_BLOCK_SIZE);
	sms4_encrypt(ctx, out, in);
	for (i = 0; i + SMS4_BLOCK_SIZE < icd_size; i += SMS4_BLOCK_SIZE) {
		memcpy(in, integrity_check_data + i, SMS4_BLOCK_SIZE);
		xor_block(in, out);
		sms4_encrypt(ctx, out, in);
	}
	/* special treatment -
	   last block might be shorter than SMS4_BLOCK_SIZE */
	if (icd_size % SMS4_BLOCK_SIZE != 0) {
		memset(in, 0, SMS4_BLOCK_SIZE);
		memcpy(in, integrity_check_data + i,
		       icd_size % SMS4_BLOCK_SIZE);
	} else {
		memcpy(in, integrity_check_data + i, SMS4_BLOCK_SIZE);
	}
	xor_block(in, out);
	sms4_encrypt(ctx, out, in);

	memcpy(mic, out, SMS4_BLOCK_SIZE);
}
